==> congruence-0.2.0.crate: congruence-0.2.0/tests/symbol_lang.rs <==
use std::{hash::Hash, rc::Rc};

use congruence::{CongruenceClosure, CongruenceState, DisjointSetForest, Language, UnionFind};
use indexmap::{map::Entry, IndexMap};

#[derive(Debug, Clone, Eq, PartialEq, Hash)]
pub enum SymbolTree {
    App(Rc<SymbolTree>, Rc<SymbolTree>),
    Pair(Rc<SymbolTree>, Rc<SymbolTree>),
    Atom(String),
}

macro_rules! sym {
    (($l:tt, $r:tt)) => {
        Rc::new(SymbolTree::App(sym!($l), sym!($r)))
    };
    ([$l:tt, $r:tt]) => {
        Rc::new(SymbolTree::Pair(sym!($l), sym!($r)))
    };
    ($e:literal) => {
        Rc::new(SymbolTree::Atom(String::from($e)))
    };
    ($v:expr) => {
        ($v).clone()
    };
}

impl SymbolTree {
    pub fn visit_children(&self, visitor: &mut impl FnMut(Rc<SymbolTree>)) {
        match self {
            SymbolTree::App(left, right) | SymbolTree::Pair(left, right) => {
                visitor(left.clone());
                visitor(right.clone());
            }
            _ => {}
        }
    }
}

#[derive(Debug, Default)]
pub struct SymbolTreeCongruence {
    intern: SymbolTreeIntern,
    cc: CongruenceClosure<Rc<SymbolTree>>,
    cs: CongruenceState,
}

impl SymbolTreeCongruence {
    #[inline]
    pub fn insert(&mut self, tree: Rc<SymbolTree>) -> (usize, bool) {
        let ix = match self.intern.table.entry(tree.clone()) {
            Entry::Occupied(o) => return (o.index(), false),
            Entry::Vacant(v) => {
                let ix = v.index();
                v.insert(());
                ix
            }
        };
        tree.visit_children(&mut |child| {
            self.insert(child);
        });
        self.cc.equation(tree, ix, &mut self.intern, &mut self.cs);
        (ix, true)
    }

    pub fn clear(&mut self) {
        self.intern.table.clear();
        self.intern.dsf.clear();
        self.cc.clear();
    }

    pub fn merge(&mut self, left: Rc<SymbolTree>, right: Rc<SymbolTree>) {
        let left_ix = self.insert(left.clone()).0;
        let right_ix = self.insert(right.clone()).0;
        self.cc
            .merge(left_ix, right_ix, &mut self.intern, &mut self.cs)
    }

    pub fn cong(&self, left: &SymbolTree, right: &SymbolTree) -> bool {
        (match (self.intern.ix(left), self.intern.ix(right)) {
            (Some(left), Some(right)) => return self.intern.dsf.node_eq(left, right),
            (None, Some(right)) => self.cc.expr_cong(left, right, &self.intern),
            (Some(left), None) => self.cc.expr_cong(right, left, &self.intern),
            (None, None) => false,
        }) || (match (left, right) {
            (SymbolTree::App(ll, lr), SymbolTree::App(rl, rr))
            | (SymbolTree::Pair(ll, lr), SymbolTree::Pair(rl, rr)) => {
                self.cong(ll, rl) && self.cong(lr, rr)
            }
            _ => left == right,
        })
    }

    pub fn cong_mut(&mut self, left: &SymbolTree, right: &SymbolTree) -> bool {
        (match (self.intern.ix(left), self.intern.ix(right)) {
            (Some(left), Some(right)) => return self.intern.dsf.node_eq_mut(left, right),
            (None, Some(right)) => self.cc.expr_cong_mut(left, right, &mut self.intern),
            (Some(left), None) => self.cc.expr_cong_mut(right, left, &mut self.intern),
            (None, None) => false,
        }) || (match (left, right) {
            (SymbolTree::App(ll, lr), SymbolTree::App(rl, rr))
            | (SymbolTree::Pair(ll, lr), SymbolTree::Pair(rl, rr)) => {
                self.cong_mut(ll, rl) && self.cong_mut(lr, rr)
            }
            _ => left == right,
        })
    }
}

#[derive(Debug, Default)]
pub struct SymbolTreeIntern {
    table: IndexMap<Rc<SymbolTree>, ()>,
    dsf: DisjointSetForest,
}

impl SymbolTreeIntern {
    #[inline]
    pub fn ix(&self, tree: &SymbolTree) -> Option<usize> {
        self.table.get_index_of(tree)
    }
}

impl Language<SymbolTreeIntern, usize> for SymbolTree {
    fn visit_deps<E>(
        &self,
        visitor: &mut impl FnMut(&SymbolTreeIntern, usize) -> Result<(), E>,
        ctx: &SymbolTreeIntern,
    ) -> Result<(), E> {
        match self {
            SymbolTree::App(left, right) => {
                if let Some(left) = ctx.ix(left) {
                    visitor(ctx, left)?;
                }
                if let Some(right) = ctx.ix(right) {
                    visitor(ctx, right)?
                }
            }
            _ => {}
        }
        Ok(())
    }

    fn visit_deps_mut<E>(
        &self,
        visitor: &mut impl FnMut(&mut SymbolTreeIntern, usize) -> Result<(), E>,
        ctx: &mut SymbolTreeIntern,
    ) -> Result<(), E> {
        match self {
            SymbolTree::App(left, right) => {
                if let Some(left) = ctx.ix(left) {
                    visitor(ctx, left)?;
                }
                if let Some(right) = ctx.ix(right) {
                    visitor(ctx, right)?
                }
            }
            _ => {}
        }
        Ok(())
    }

    fn hash_mod(
        &self,
        hasher: &mut impl std::hash::Hasher,
        map: &mut impl FnMut(&SymbolTreeIntern, usize) -> usize,
        ctx: &SymbolTreeIntern,
    ) {
        std::mem::discriminant(self).hash(hasher);
        match self {
            SymbolTree::App(left, right) | SymbolTree::Pair(left, right) => {
                if let Some(left) = ctx.ix(left) {
                    map(ctx, left).hash(hasher);
                } else {
                    left.hash(hasher);
                }
                if let Some(right) = ctx.ix(right) {
                    map(ctx, right).hash(hasher);
                } else {
                    right.hash(hasher);
                }
            }
            SymbolTree::Atom(atom) => atom.hash(hasher),
        }
    }

    fn hash_mod_mut(
        &self,
        hasher: &mut impl std::hash::Hasher,
        map: &mut impl FnMut(&mut SymbolTreeIntern, usize) -> usize,
        ctx: &mut SymbolTreeIntern,
    ) {
        std::mem::discriminant(self).hash(hasher);
        match self {
            SymbolTree::App(left, right) | SymbolTree::Pair(left, right) => {
                if let Some(left) = ctx.ix(left) {
                    map(ctx, left).hash(hasher);
                } else {
                    left.hash(hasher);
                }
                if let Some(right) = ctx.ix(right) {
                    map(ctx, right).hash(hasher);
                } else {
                    right.hash(hasher);
                }
            }
            SymbolTree::Atom(atom) => atom.hash(hasher),
        }
    }

    fn eq_mod(
        &self,
        other: &Self,
        map: &mut impl FnMut(&SymbolTreeIntern, usize, usize) -> bool,
        ctx: &SymbolTreeIntern,
    ) -> bool {
        match (self, other) {
            (SymbolTree::App(this_l, this_r), SymbolTree::App(other_l, other_r))
            | (SymbolTree::Pair(this_l, this_r), SymbolTree::Pair(other_l, other_r)) => {
                (if let (Some(this_l), Some(other_l)) = (ctx.ix(this_l), ctx.ix(other_l)) {
                    map(ctx, this_l, other_l)
                } else {
                    this_l == other_l
                }) && (if let (Some(this_r), Some(other_r)) = (ctx.ix(this_r), ctx.ix(other_r)) {
                    map(ctx, this_r, other_r)
                } else {
                    this_r == other_r
                })
            }
            (SymbolTree::Atom(this), SymbolTree::Atom(other)) => this == other,
            _ => false,
        }
    }

    fn eq_mod_mut(
        &self,
        other: &Self,
        map: &mut impl FnMut(&mut SymbolTreeIntern, usize, usize) -> bool,
        ctx: &mut SymbolTreeIntern,
    ) -> bool {
        match (self, other) {
            (SymbolTree::App(this_l, this_r), SymbolTree::App(other_l, other_r))
            | (SymbolTree::Pair(this_l, this_r), SymbolTree::Pair(other_l, other_r)) => {
                (if let (Some(this_l), Some(other_l)) = (ctx.ix(this_l), ctx.ix(other_l)) {
                    map(ctx, this_l, other_l)
                } else {
                    this_l == other_l
                }) && (if let (Some(this_r), Some(other_r)) = (ctx.ix(this_r), ctx.ix(other_r)) {
                    map(ctx, this_r, other_r)
                } else {
                    this_r == other_r
                })
            }
            (SymbolTree::Atom(this), SymbolTree::Atom(other)) => this == other,
            _ => false,
        }
    }
}

impl UnionFind for SymbolTreeIntern {
    fn union_find(&mut self, left: usize, right: usize) -> usize {
        self.dsf.union_find(left, right)
    }

    fn pre_union_find(&mut self, left: usize, right: usize) -> usize {
        self.dsf.pre_union_find(left, right)
    }

    fn union(&mut self, left: usize, right: usize) {
        self.dsf.union(left, right)
    }

    fn find(&self, node: usize) -> usize {
        self.dsf.find(node)
    }

    fn node_eq(&self, left: usize, right: usize) -> bool {
        self.dsf.node_eq(left, right)
    }

    fn find_mut(&mut self, node: usize) -> usize {
        self.dsf.find_mut(node)
    }

    fn node_eq_mut(&mut self, left: usize, right: usize) -> bool {
        self.dsf.node_eq_mut(left, right)
    }
}

#[test]
fn basic_test() {
    let mut cc = SymbolTreeCongruence::default();
    let f = sym!("f");
    let g = sym!("g");
    let x = sym!("x");
    let y = sym!("y");
    let f_x = sym!((f, x));
    let g_y = sym!((g, y));
    let g_x = sym!((g, x));
    let f_x_y = sym!((f_x, y));
    let f_x_y_p = sym!([f_x, y]);
    assert!(!cc.cong_mut(&f_x, &g));
    assert!(!cc.cong_mut(&g_y, &y));
    assert!(!cc.cong_mut(&g_x, &x));
    assert!(!cc.cong_mut(&f_x_y, &y));
    cc.merge(f_x.clone(), g.clone());
    assert!(cc.cong_mut(&f_x, &g));
    assert!(!cc.cong_mut(&g_y, &y));
    cc.merge(g_y.clone(), y.clone());
    assert!(cc.cong_mut(&f_x, &g));
    assert!(cc.cong_mut(&g_y, &y));
    assert!(!cc.cong_mut(&g_x, &x));
    assert!(cc.cong_mut(&f_x_y, &y));
    assert!(!cc.cong_mut(&f_x_y_p, &y));
    cc.insert(f_x_y.clone());
    assert!(!cc.cong_mut(&g_x, &x));
    assert!(cc.cong_mut(&f_x_y, &y));
    assert!(!cc.cong_mut(&f_x_y_p, &y));
}

#[test]
fn g_x_tower() {
    let mut cc = SymbolTreeCongruence::default();
    let g = sym!("g");
    let x = sym!("x");
    let app_1 = sym!((g, x));
    let app_2 = sym!((g, (g, x)));
    let app_3 = sym!((g, (g, (g, x))));
    let app_5 = sym!((g, (g, (g, (g, (g, x))))));
    cc.merge(app_3.clone(), x.clone());
    assert!(!cc.cong(&app_1, &x));
    assert!(cc.cong(&app_3, &x));
    assert!(!cc.cong(&app_5, &x));
    assert!(cc.cong(&app_5, &app_2));
    cc.merge(app_5.clone(), x.clone());
    assert!(cc.cong(&app_1, &x));
    assert!(cc.cong(&app_2, &x));
    assert!(cc.cong(&app_3, &x));
    assert!(cc.cong(&app_5, &x));
    cc.clear();
    cc.merge(app_5.clone(), x.clone());
    assert!(!cc.cong(&app_1, &x));
    assert!(!cc.cong(&app_3, &x));
    assert!(cc.cong(&app_5, &x));
    assert!(!cc.cong(&app_5, &app_2));
    cc.merge(app_3.clone(), x.clone());
    assert!(cc.cong(&app_1, &x));
    assert!(cc.cong(&app_2, &x));
    assert!(cc.cong(&app_3, &x));
    assert!(cc.cong(&app_5, &x));
}

==> congruence-0.3.0.crate: congruence-0.3.0/tests/symbol_lang.rs <==
use std::{hash::{Hash, Hasher}, rc::Rc};

use congruence::{CongruenceClosure, CongruenceState, DisjointSetForest, Language, UnionFind};
use indexmap::{map::Entry, IndexMap};

#[derive(Debug, Clone, Eq, PartialEq, Hash)]
pub enum SymbolTree {
    App(Rc<SymbolTree>, Rc<SymbolTree>),
    Pair(Rc<SymbolTree>, Rc<SymbolTree>),
    Atom(String),
}

macro_rules! sym {
    (($l:tt, $r:tt)) => {
        Rc::new(SymbolTree::App(sym!($l), sym!($r)))
    };
    ([$l:tt, $r:tt]) => {
        Rc::new(SymbolTree::Pair(sym!($l), sym!($r)))
    };
    ($e:literal) => {
        Rc::new(SymbolTree::Atom(String::from($e)))
    };
    ($v:expr) => {
        ($v).clone()
    };
}

impl SymbolTree {
    pub fn visit_children(&self, visitor: &mut impl FnMut(Rc<SymbolTree>)) {
        match self {
            SymbolTree::App(left, right) | SymbolTree::Pair(left, right) => {
                visitor(left.clone());
                visitor(right.clone());
            }
            _ => {}
        }
    }
}

#[derive(Debug, Default)]
pub struct SymbolTreeCongruence {
    intern: SymbolTreeIntern,
    cc: CongruenceClosure<Rc<SymbolTree>>,
    cs: CongruenceState,
}

impl SymbolTreeCongruence {
    #[inline]
    pub fn insert(&mut self, tree: Rc<SymbolTree>) -> (usize, bool) {
        let ix = match self.intern.table.entry(tree.clone()) {
            Entry::Occupied(o) => return (o.index(), false),
            Entry::Vacant(v) => {
                let ix = v.index();
                v.insert(());
                ix
            }
        };
        tree.visit_children(&mut |child| {
            self.insert(child);
        });
        self.cc.equation(tree, ix, &mut self.intern, &mut self.cs);
        (ix, true)
    }

    pub fn clear(&mut self) {
        self.intern.table.clear();
        self.intern.dsf.clear();
        self.cc.clear();
    }

    pub fn merge(&mut self, left: Rc<SymbolTree>, right: Rc<SymbolTree>) {
        let left_ix = self.insert(left.clone()).0;
        let right_ix = self.insert(right.clone()).0;
        self.cc
            .merge(left_ix, right_ix, &mut self.intern, &mut self.cs)
    }

    pub fn cong(&self, left: &SymbolTree, right: &SymbolTree) -> bool {
        (match (self.intern.ix(left), self.intern.ix(right)) {
            (Some(left), Some(right)) => return self.intern.dsf.node_eq(left, right),
            (None, Some(right)) => self.cc.expr_cong(left, right, &self.intern),
            (Some(left), None) => self.cc.expr_cong(right, left, &self.intern),
            (None, None) => false,
        }) || (match (left, right) {
            (SymbolTree::App(ll, lr), SymbolTree::App(rl, rr))
            | (SymbolTree::Pair(ll, lr), SymbolTree::Pair(rl, rr)) => {
                self.cong(ll, rl) && self.cong(lr, rr)
            }
            _ => left == right,
        })
    }

    pub fn cong_mut(&mut self, left: &SymbolTree, right: &SymbolTree) -> bool {
        (match (self.intern.ix(left), self.intern.ix(right)) {
            (Some(left), Some(right)) => return self.intern.dsf.node_eq_mut(left, right),
            (None, Some(right)) => self.cc.expr_cong_mut(left, right, &mut self.intern),
            (Some(left), None) => self.cc.expr_cong_mut(right, left, &mut self.intern),
            (None, None) => false,
        }) || (match (left, right) {
            (SymbolTree::App(ll, lr), SymbolTree::App(rl, rr))
            | (SymbolTree::Pair(ll, lr), SymbolTree::Pair(rl, rr)) => {
                self.cong_mut(ll, rl) && self.cong_mut(lr, rr)
            }
            _ => left == right,
        })
    }
}

#[derive(Debug, Default)]
pub struct SymbolTreeIntern {
    table: IndexMap<Rc<SymbolTree>, ()>,
    dsf: DisjointSetForest,
}

impl SymbolTreeIntern {
    #[inline]
    pub fn ix(&self, tree: &SymbolTree) -> Option<usize> {
        self.table.get_index_of(tree)
    }
}

impl Language<SymbolTreeIntern, usize> for SymbolTree {
    fn visit_deps<E>(
        &self,
        visitor: &mut impl FnMut(&SymbolTreeIntern, usize) -> Result<(), E>,
        ctx: &SymbolTreeIntern,
    ) -> Result<(), E> {
        match self {
            SymbolTree::App(left, right) => {
                if let Some(left) = ctx.ix(left) {
                    visitor(ctx, left)?;
                }
                if let Some(right) = ctx.ix(right) {
                    visitor(ctx, right)?
                }
            }
            _ => {}
        }
        Ok(())
    }

    fn visit_deps_mut<E>(
        &self,
        visitor: &mut impl FnMut(&mut SymbolTreeIntern, usize) -> Result<(), E>,
        ctx: &mut SymbolTreeIntern,
    ) -> Result<(), E> {
        match self {
            SymbolTree::App(left, right) => {
                if let Some(left) = ctx.ix(left) {
                    visitor(ctx, left)?;
                }
                if let Some(right) = ctx.ix(right) {
                    visitor(ctx, right)?
                }
            }
            _ => {}
        }
        Ok(())
    }

    fn hash_mod<H: Hasher>(
        &self,
        hasher: &mut H,
        map: &mut impl FnMut(&mut H, &SymbolTreeIntern, usize),
        ctx: &SymbolTreeIntern,
    ) {
        std::mem::discriminant(self).hash(hasher);
        match self {
            SymbolTree::App(left, right) | SymbolTree::Pair(left, right) => {
                if let Some(left) = ctx.ix(left) {
                    map(hasher, ctx, left);
                } else {
                    left.hash(hasher);
                }
                if let Some(right) = ctx.ix(right) {
                    map(hasher, ctx, right);
                } else {
                    right.hash(hasher);
                }
            }
            SymbolTree::Atom(atom) => atom.hash(hasher),
        }
    }

    fn hash_mod_mut<H: Hasher>(
        &self,
        hasher: &mut H,
        map: &mut impl FnMut(&mut H, &mut SymbolTreeIntern, usize),
        ctx: &mut SymbolTreeIntern,
    ) {
        std::mem::discriminant(self).hash(hasher);
        match self {
            SymbolTree::App(left, right) | SymbolTree::Pair(left, right) => {
                if let Some(left) = ctx.ix(left) {
                    map(hasher, ctx, left);
                } else {
                    left.hash(hasher);
                }
                if let Some(right) = ctx.ix(right) {
                    map(hasher, ctx, right);
                } else {
                    right.hash(hasher);
                }
            }
            SymbolTree::Atom(atom) => atom.hash(hasher),
        }
    }

    fn eq_mod(
        &self,
        other: &Self,
        map: &mut impl FnMut(&SymbolTreeIntern, usize, usize) -> bool,
        ctx: &SymbolTreeIntern,
    ) -> bool {
        match (self, other) {
            (SymbolTree::App(this_l, this_r), SymbolTree::App(other_l, other_r))
            | (SymbolTree::Pair(this_l, this_r), SymbolTree::Pair(other_l, other_r)) => {
                (if let (Some(this_l), Some(other_l)) = (ctx.ix(this_l), ctx.ix(other_l)) {
                    map(ctx, this_l, other_l)
                } else {
                    this_l == other_l
                }) && (if let (Some(this_r), Some(other_r)) = (ctx.ix(this_r), ctx.ix(other_r)) {
                    map(ctx, this_r, other_r)
                } else {
                    this_r == other_r
                })
            }
            (SymbolTree::Atom(this), SymbolTree::Atom(other)) => this == other,
            _ => false,
        }
    }

    fn eq_mod_mut(
        &self,
        other: &Self,
        map: &mut impl FnMut(&mut SymbolTreeIntern, usize, usize) -> bool,
        ctx: &mut SymbolTreeIntern,
    ) -> bool {
        match (self, other) {
            (SymbolTree::App(this_l, this_r), SymbolTree::App(other_l, other_r))
            | (SymbolTree::Pair(this_l, this_r), SymbolTree::Pair(other_l, other_r)) => {
                (if let (Some(this_l), Some(other_l)) = (ctx.ix(this_l), ctx.ix(other_l)) {
                    map(ctx, this_l, other_l)
                } else {
                    this_l == other_l
                }) && (if let (Some(this_r), Some(other_r)) = (ctx.ix(this_r), ctx.ix(other_r)) {
                    map(ctx, this_r, other_r)
                } else {
                    this_r == other_r
                })
            }
            (SymbolTree::Atom(this), SymbolTree::Atom(other)) => this == other,
            _ => false,
        }
    }
}

impl UnionFind for SymbolTreeIntern {
    fn union_find(&mut self, left: usize, right: usize) -> usize {
        self.dsf.union_find(left, right)
    }

    fn pre_union_find(&mut self, left: usize, right: usize) -> usize {
        self.dsf.pre_union_find(left, right)
    }

    fn union(&mut self, left: usize, right: usize) {
        self.dsf.union(left, right)
    }

    fn find(&self, node: usize) -> usize {
        self.dsf.find(node)
    }

    fn node_eq(&self, left: usize, right: usize) -> bool {
        self.dsf.node_eq(left, right)
    }

    fn find_mut(&mut self, node: usize) -> usize {
        self.dsf.find_mut(node)
    }

    fn node_eq_mut(&mut self, left: usize, right: usize) -> bool {
        self.dsf.node_eq_mut(left, right)
    }
}

#[test]
fn basic_test() {
    let mut cc = SymbolTreeCongruence::default();
    let f = sym!("f");
    let g = sym!("g");
    let x = sym!("x");
    let y = sym!("y");
    let f_x = sym!((f, x));
    let g_y = sym!((g, y));
    let g_x = sym!((g, x));
    let f_x_y = sym!((f_x, y));
    let f_x_y_p = sym!([f_x, y]);
    assert!(!cc.cong_mut(&f_x, &g));
    assert!(!cc.cong_mut(&g_y, &y));
    assert!(!cc.cong_mut(&g_x, &x));
    assert!(!cc.cong_mut(&f_x_y, &y));
    cc.merge(f_x.clone(), g.clone());
    assert!(cc.cong_mut(&f_x, &g));
    assert!(!cc.cong_mut(&g_y, &y));
    cc.merge(g_y.clone(), y.clone());
    assert!(cc.cong_mut(&f_x, &g));
    assert!(cc.cong_mut(&g_y, &y));
    assert!(!cc.cong_mut(&g_x, &x));
    assert!(cc.cong_mut(&f_x_y, &y));
    assert!(!cc.cong_mut(&f_x_y_p, &y));
    cc.insert(f_x_y.clone());
    assert!(!cc.cong_mut(&g_x, &x));
    assert!(cc.cong_mut(&f_x_y, &y));
    assert!(!cc.cong_mut(&f_x_y_p, &y));
}

#[test]
fn g_x_tower() {
    let mut cc = SymbolTreeCongruence::default();
    let g = sym!("g");
    let x = sym!("x");
    let app_1 = sym!((g, x));
    let app_2 = sym!((g, (g, x)));
    let app_3 = sym!((g, (g, (g, x))));
    let app_5 = sym!((g, (g, (g, (g, (g, x))))));
    cc.merge(app_3.clone(), x.clone());
    assert!(!cc.cong(&app_1, &x));
    assert!(cc.cong(&app_3, &x));
    assert!(!cc.cong(&app_5, &x));
    assert!(cc.cong(&app_5, &app_2));
    cc.merge(app_5.clone(), x.clone());
    assert!(cc.cong(&app_1, &x));
    assert!(cc.cong(&app_2, &x));
    assert!(cc.cong(&app_3, &x));
    assert!(cc.cong(&app_5, &x));
    cc.clear();
    cc.merge(app_5.clone(), x.clone());
    assert!(!cc.cong(&app_1, &x));
    assert!(!cc.cong(&app_3, &x));
    assert!(cc.cong(&app_5, &x));
    assert!(!cc.cong(&app_5, &app_2));
    cc.merge(app_3.clone(), x.clone());
    assert!(cc.cong(&app_1, &x));
    assert!(cc.cong(&app_2, &x));
    assert!(cc.cong(&app_3, &x));
    assert!(cc.cong(&app_5, &x));
}
